/*
**	mod_snmp.c	- SNMP Interface module
**
**
** Copyright (c) 1997  Hughes Technologies Pty Ltd
**
** Permission to use, copy, and distribute for non-commercial purposes,
** is hereby granted without fee, providing that the above copyright
** notice appear in all copies and that both the copyright notice and this
** permission notice appear in supporting documentation.
**
** The software may be modified for your own purposes, but modified versions
** may not be distributed.
**
** This software is provided "as is" without any expressed or implied warranty.
**
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <netinet/in.h>
#include <netdb.h>

#include <ctype.h>
#include <sys/time.h>
#include <sys/types.h>

#include <common/portability.h>
#include <lite.h>

#include "snmp.h"

#include "mod_snmp.h"


/*
** Internal session handle structure
*/

typedef struct snmpfd_s {
	int	fd;
	snmp 	*session;
	struct snmpfd_s *next;
} snmpfd_t;



/*
** Globals
*/

extern  sym_t   *externReturn;
static	snmpfd_t *snmpFD = NULL;
static	int	curFD = 1;


/****************************************************************************
** 	_initSnmpModule
**
**	Purpose	: initialise the SNMP library
**	Args	: None
**	Returns	: Nothing
**	Notes	: init_mib() is slow under cmu 2.x.  Not sure why!
*/

void initSnmpModule()
{
	if (snmpInit() < 0)
	{
		runError(snmpError);
		exit(1);
	}
}







/****************************************************************************
** 	_snmpFindSession
**
**	Purpose	: Locate a session struct given a session handle
**	Args	: fd (session handle)
**	Returns	: pointer to session struct
**	Notes	: 
*/

static snmpfd_t *snmpFindSession(fd)
	int	fd;
{
	snmpfd_t *cur;

	cur = snmpFD;
	while(cur)
	{
		if (cur->fd == fd)
			break;
		cur = cur->next;
	}
	return(cur);
}





/****************************************************************************
** 	_snmpOpen
**
**	Purpose	: Open an SNMP session with a host
**	Args	: hostname and community
**	Returns	: Nothing
**	Notes	: 
*/

void doSnmpOpen(params)
	plist_t	*params;
{
	char	*host,
		*community;
	snmpfd_t *new;
	snmp	*session;
	int	result;

	host = (char *)params->sym->val;
	community = (char *)params->next->sym->val;
	session = snmpOpen(host,community,SNMP_DEF_TIMEOUT,SNMP_DEF_RETRIES);
	if (!session)
	{
		externReturn = createIntSymbol(-1);
		setError(snmpError);
		return;
	}

	curFD++;
	new = (snmpfd_t *)malloc(sizeof(snmpfd_t));
	if (!new)
		runError("Out of memory");
	result = new->fd = curFD;
	new->session = session;
	new->next = snmpFD;
	snmpFD = new;
	externReturn = createIntSymbol(result);
}





/****************************************************************************
** 	_snmpClose
**
**	Purpose	: Shut down an snmp session
**	Args	: (int) session descriptor
**	Returns	: Nothing
**	Notes	: Releases the session descriptor and associated data
*/

void doSnmpClose(params)
	plist_t	*params;
{
	int	fd;
	snmpfd_t *cur,
		 *prev;

	fd = (int) * (int *)params->sym->val;
	cur = snmpFD;
	prev = NULL;
	while(cur)
	{
		if (cur->fd == fd)
		{
			snmpClose(cur->session);
			if (!prev)
				snmpFD = snmpFD->next;
			else
				prev->next = cur->next;
			free(cur);
			cur = NULL;
			return;
		}
		prev = cur;
		cur = cur->next;
	}
	if (!cur)
	{
         	setError("$errmsg","Bad SNMP descriptor");
                return;
	}
}



/****************************************************************************
** 	_snmpGet
**
**	Purpose	: GET/GETNEXT engine
**	Args	: operator (GET OR GETNEXT), fd and variables
**	Returns	: Nothing
**	Notes	: 
*/

void doSnmpGet(params)
	plist_t	*params;
{
	char	*var,
		resBuf[10240];
	int	fd,
		numVars,
		count;
	snmpfd_t *sfd;
	sym_t	*array,
		*tmp,
		*resSym;
	snmpvar	vars[64];


	fd = (int) * (int *)params->sym->val;
	numVars = symGetNumArrayElements(params->next->sym);

	array = (sym_t *)createArray();
	count = 0;
	while(count < numVars)
	{
		tmp = symGetArrayElement(params->next->sym,count);
		var = tmp->val;
		if (snmpCheckVariable(var) < 0)
		{
			setError(snmpError);
			symFreeSymbol(array);
			return;
		}
		snmpAddVariable(var,&vars[count]);
		count++;
	}

	sfd = snmpFindSession(fd);
	if (!sfd)
	{
         	setError("Bad SNMP descriptor");
		symFreeSymbol(array);
                return;
	}

	if (snmpGet(sfd->session, vars, numVars) < 0)
	{
		setError(snmpError);
		symFreeSymbol(array);
		return;
	}
	
	count = 0;
	while(count < numVars)
	{
		sprintf(resBuf,"%s%s%s",vars[count].textID, LITE_BRK,
			vars[count].textVal);
		resSym = (sym_t *)createCharSymbol(resBuf);
		symSetArrayElement(array,count++,resSym);

	}
	externReturn = array;
}






/****************************************************************************
** 	_snmpGetNext
**
**	Purpose	: GET/GETNEXT engine
**	Args	: operator (GET OR GETNEXT), fd and variables
**	Returns	: Nothing
**	Notes	: 
*/

void doSnmpGetNext(params)
	plist_t	*params;
{
	char	*var,
		resBuf[10240];
	int	fd,
		numVars,
		count;
	snmpfd_t *sfd;
	sym_t	*array,
		*tmp,
		*resSym;
	snmpvar	vars[64];


	fd = (int) * (int *)params->sym->val;
	numVars = symGetNumArrayElements(params->next->sym);

	array = (sym_t *)createArray();
	count = 0;
	while(count < numVars)
	{
		tmp = symGetArrayElement(params->next->sym,count);
		var = tmp->val;
		if (snmpCheckVariable(var) < 0)
		{
			setError(snmpError);
			symFreeSymbol(array);
			return;
		}
		snmpAddVariable(var,&vars[count]);
		count++;
	}

	sfd = snmpFindSession(fd);
	if (!sfd)
	{
         	setError("Bad SNMP descriptor");
		symFreeSymbol(array);
                return;
	}

	if (snmpGetNext(sfd->session, vars, numVars) < 0)
	{
		setError(snmpError);
		symFreeSymbol(array);
		return;
	}
	
	count = 0;
	while(count < numVars)
	{
		sprintf(resBuf,"%s%s%s",vars[count].textID, LITE_BRK,
			vars[count].textVal);
		resSym = (sym_t *)createCharSymbol(resBuf);
		symSetArrayElement(array,count++,resSym);

	}
	externReturn = array;
}







int init_mod(modPtr)
        void    *modPtr;
{
	initSnmpModule();
        if (modLoadModuleFunctions(modPtr, snmp_efuncts) < 0)
        {
                return(-1);
        }
	return(0);
}

